/* main.c - main file for the rndtool
 *
 * Copyright (C) 2002  Jochen Voss.  */

static const  char  rcsid[] = "$Id: main.c,v 1.3 2002/10/13 16:00:37 voss Exp $";

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>

#include <jvrand.h>

#include "rndlib.h"
#include "readlines.h"


static void
scan_arg_int (const char *arg, long int *res)
{
  char *tail;
  long int  val;
	
  errno = 0;
  val = strtol (arg, &tail, 10);
  if (errno == ERANGE && *tail == '\0') {
    fatal ("invalid number \"%s\"\n", arg);
  } else if (errno || tail == arg || *tail != '\0') {
    fatal ("malformed number \"%s\"\n", arg);
  } else {
    *res = val;
  }
}

static void
scan_arg_double (const char *arg, double *res)
{
  char *tail;
  double  val;
	
  errno = 0;
  val = strtod (arg, &tail);
  if (errno == ERANGE && *tail == '\0') {
    fatal ("invalid number \"%s\"\n", arg);
  } else if (errno || tail == arg || *tail != '\0') {
    fatal ("malformed number \"%s\"\n", arg);
  } else {
    *res = val;
  }
}

/**********************************************************************
 * shuffle lines
 */

struct file_buffer  buffer;

static void
shuffle_lines (void)
{
  long int  i;
  
  buffer.used = 0;
  buffer.slots = 100;
  buffer.data = xmalloc (buffer.slots*sizeof(struct line));

  yylex ();

  jvrand_shuffle (buffer.data, buffer.used, sizeof(struct line));
  
  for (i=0; i<buffer.used; ++i) {
    size_t  res;
    
    res = fwrite (buffer.data[i].data, buffer.data[i].size, 1, stdout);
    if (res != 1)  fatal ("write failed: %s", strerror (errno));
    putchar ('\n');
  }
}

static void
choose_lines (long int count)
{
  long int  i;
  
  buffer.used = 0;
  buffer.slots = 100;
  buffer.data = xmalloc (buffer.slots*sizeof(struct line));

  yylex ();

  jvrand_shuffle (buffer.data, buffer.used, sizeof(struct line));
  
  for (i=0; i<count; ++i) {
    size_t  res;
    int  k;

    k = jvrand_discrete (0, buffer.used);
    res = fwrite (buffer.data[k].data, buffer.data[k].size, 1, stdout);
    if (res != 1)  fatal ("write failed: %s", strerror (errno));
    putchar ('\n');
  }
}

/**********************************************************************
 * main program
 */

int
main (int argc, char **argv)
{
  int  error_flag = 0;		/* Show usage and exit.  */
  int  version_flag = 0;	/* Show version information (-V) */
  int  help_flag = 0;		/* Show help and exit (-h) */
  int  sum_flag = 0;		/* Summation mode (-S) */
  int  cumsum_flag = 0;		/* Cumulative summation mode (-C) */
  int  zero_flag = 0;		/* Start with 0 for cumsum mode (-Z) */
  int  modes [256];		/* Flags for operation modes */
  long int  count = 1;
  long int  arg1 = 0;
  struct voption  option_tab [] = {
    { "choose-arg", 'a', modes+'a', 0, NULL,
        "emit one of the command line arguments" },
    { "choose-line", 'l', modes+'l', 0, NULL,
        "emit one randomly choosen line from stdin" },
    { "choose-some-args", 'k', modes+'k', 1, "N",
        "emit N of the command line arguments" },
    { "count", 'N', NULL, 1, "N",
        "emit N independent identically distributed values" },
    { "cumsum", 'C', &cumsum_flag, 0, NULL,
        "emit the cummulative sum of the results" },
    { "help", 'h', &help_flag, 0, NULL,
        "show this message and exit" },
    { "random-discrete", 'd', modes+'d', 0, NULL,
        "emit a random integer (default)" },
    { "random-normal", 'n', modes+'n', 0, NULL,
        "emit a normally distributed random value" },
    { "shuffle-args", 'r', modes+'r', 0, NULL,
        "shuffle the command line arguments" },
    { "shuffle-lines", 's', modes+'s', 0, NULL,
        "shuffle the lines from stdin" },
    { "sum", 'S', &sum_flag, 0, NULL,
        "emit the sum of the results" },
    { "version", 'V', &version_flag, 0, NULL,
        "show the random-utils version information" },
    { "zero", 'Z', &zero_flag, 0, NULL,
        "emit a leading 0 (for cumsum mode)" },
    { NULL, '\0', NULL, 0, NULL, NULL }
  };
  char mode = 'd';
  double  x = 0, s = 0;
  long int i;
  int n;

  vinit (argc, argv);
  jvrand_init ();

  for (i=0; i<256; ++i)  modes[i] = 0;
  while (! error_flag) {
    const char *argptr;
    int  c = vgetopt (option_tab, &argptr, 0);

    if (c == -1)  break;
    switch (c) {
    case 'k':
      scan_arg_int (argptr, &arg1);
      break;
    case 'N':
      scan_arg_int (argptr, &count);
      break;
    case 0:
      error ("unkown option '%s'", argptr);
      error_flag = 1;
      break;
    default:
      assert (0);
    }
  }

  if (error_flag || help_flag) {
    printf ("Usage: %s [options]\n", myname);
    if (error_flag) {
      printf ("Try `%s --help' for more information.\n", myname);
      if (! help_flag)  exit (EXIT_FAILURE);
    }
    putchar ('\n');
    show_options (option_tab);
    putchar ('\n');
    puts ("Please report bugs to <" PACKAGE_BUGREPORT ">");
    exit (error_flag ? EXIT_FAILURE : EXIT_SUCCESS);
  }
  if (version_flag) {
    puts ("rnd-tools " VERSION);
    puts ("Copyright (C) 1999, 2002 Jochen Voss");
    puts ("\
Random-utils comes with NO WARRANTY, to the extent permitted by law.\n\
You may redistribute copies of random-utils under the terms of the\n\
GNU General Public License.  For more information about these matters,\n\
see the file named COPYING from the source code distribution.");
    exit (0);
  }

  n = 0;
  for (i=0; i<256; ++i)  n += modes[i];
  if (n>1)  fatal ("conficting modes selected");
  for (i=0; i<256; ++i) {
    if (modes[i]) {
      mode = i;
      break;
    }
  }

  if (count<0)  fatal ("negative repeat count");
  if (count != 1 && mode != 'a' && mode != 'd' && mode != 'n' && mode != 'l') {
    fatal("repeat count not allowed for this mode");
  }

  if (sum_flag && cumsum_flag) {
    fatal ("-C and -S cannot be combined");
  }
  if ((sum_flag || cumsum_flag) && mode != 'd' && mode != 'n') {
    fatal ("summation not allowed for this mode");
  }
  if (zero_flag) {
    if (! cumsum_flag)  fatal ("-Z can only be used in cumsum mode");
    puts ("0");
  }

  /* Modes of operation start here */
  
  if (mode == 'a') {		/* --choose-arg */
    for (i=0; i<count; ++i) {
      puts (xargv[jvrand_discrete (1, xargc)]);
    }
  } else if (mode == 'k') {	/* --choose-some-args */
    if (arg1<0 || arg1>=xargc) {
      fatal ("invalid argument count N=%ld (must be 0<=N<=%d)", arg1, xargc-1);
    }
    for (i=1; i<xargc && arg1>0; ++i) {
      if (jvrand_discrete (0, xargc-i) < arg1 ) {
	--arg1;
	puts (xargv[i]);
      }
    }
  } else if (mode == 'd') {	/* --random-discrete */
    long int  min = 0;
    long int  max = 65536;

    switch (xargc) {
    case 3:
      scan_arg_int (xargv[1], &min);
      scan_arg_int (xargv[2], &max);
      break;
    case 2:
      scan_arg_int (xargv[1], &max);
      break;
    case 1:
      break;
    default:
      fatal ("only 0, 1, or 2 arguments are allowed in random-discrete mode");
    }
    if (! (min < max))  fatal ("invalid range %ld, ..., %ld", min, max-1);

    for (i=0; i<count; ++i) {
      x = jvrand_discrete (min, max);
      s += x;
      if (!sum_flag) {
	printf ("%g\n", cumsum_flag ? s : x);
      }
    }
    if (sum_flag)  printf ("%g\n", s);
  } else if (mode == 'l') {	/* --choose-line */
    choose_lines (count);
  } else if (mode == 'n') {	/* --random-normal */
    double  m = 0;
    double  v = 1;

    switch (xargc) {
    case 3:
      scan_arg_double (xargv[1], &m);
      scan_arg_double (xargv[2], &v);
      break;
    case 2:
      scan_arg_double (xargv[1], &m);
      break;
    case 1:
      break;
    default:
      fatal ("only 0, 1, or 2 arguments are allowed in random-normal mode");
    }
    if (v<=0)  fatal ("invalid variance %g (must be positive)", v);

    for (i=0; i<count; ++i) {
      x = jvrand_normal (m, v);
      s += x;
      if (!sum_flag) {
	printf ("%g\n", cumsum_flag ? s : x);
      }
    }
    if (sum_flag)  printf ("%g\n", s);
  } else if (mode == 'r') {	/* --shuffle-args */
    jvrand_shuffle (xargv+1, xargc-1, sizeof(char *));
    for (i=1; i<xargc; ++i)  printf ("%s\n", xargv[i]);
  } else if (mode == 's') {	/* --shuffle-lines */
    shuffle_lines ();
  } else {
    assert (0);
  }
  
  vexit ();
  return  0;
}
